from enum import *
import sys
from math import inf
import math

FLAGBUNTIL=0

# classe pour representer une structure de Kripke pointee.
# Les attributs sont :
#    start       : l'etat initial, un string-tuple.
#    label       : un dictionnaire, qui a un etat (string-tuple) de la structure,
#                  associe l'ensemble des propositions atomiques (string) vraies dans cet etat.
#    act         : une liste d'adjacence : un dictionnaire, qui a un etat (string-tuple) de la structure,
#                  associe un dictionnaire, qui a une action (string) de la structure,
#                  associe l'ensemble des successeurs de l'etat pour cette action.
#                  Si l'action vaut None, les successeurs n'ont pas d'action.
#    succ        : une liste d'adjacence, qui a un etat,
#                  associe ses successeurs pour une action quelconque.
#                  En effet, l'action n'a d'importance que pour la synchronisation,
#                  et on se servira de succ pour tout le reste.
#                  Parcourir act pour le model-checking nous fera eventuellement parcourir
#                  le meme etat plusieurs fois.
#    accessibles : un dictionnaire, qui a un etat (string-tuple) s de la structure,
#                  associe la liste des etats accessibles par s.

# On remarque qu'on utilise surtout des ensembles et non des listes.
# En effet, on veut eviter des doublons crees par un utilisateur negligeant
# ou bien par deux actions differentes qui pointent vers le meme successeur.

# succ et accessibles se deduisent directement de act et start.
# Pour construire une structure, on n'a donc besoin que de start, label et act.
# On a une methode update (appellee lors de la construction) qui calcule alors succ et accessibles.
# L'idee est que l'utilisateur peut apporter des modifications a start, label et act dans l'interpreteur.
# Quand ces modifications seront finies, il devra appeler update.

# On peut lire une structure a partir d'un fichier. Le format est le suivant :

#    On entre la ligne "states\n" pour dire qu'on annonce les labels,
#    et "transitions\n" pour dire qu'on annonce les successeurs.
#    On entre la ligne "end\n" pour dire qu'on a fini de decrire la structure.

#    Dans la partie label, on ecrit q:{p_1,p_2,...,p_n} pour dire que les
#    p_i sont des propositions atomiques satisfaites par q.
#    De telles annonces peuvent etre separees par des virgules ou des sauts de ligne.

#    Dans la partie successeurs, on ecrit q-action>succ1|succ2|...|succn
#    pour dire que les succi sont les successeurs de q pour l'action action.
#    Si on ne veut pas preciser d'action, on ecrit q->succi.

#    Dans les deux parties, on peut separer ou aerer de telles declarations avec des espaces.
#    On peut aussi separer deux declarations successives par une virgule et/ou un saut de ligne.

#    Le premier etat annonce ainsi sera l'etat initial de la structure.

class Krip :
    # Met a jour succ et accessibles.
    def update(self,FLAGBUNTIL =0) :
        self.succ={}
        for s in self.label :
            self.succ[s]=set([])
            for action in self.act[s] :
                for succ in self.act[s][action] :
                    self.succ[s].add(succ)

        # On va maintenant calculer les accessibles de chaque sommet.
        # On utilise l'algorithme de Floyd-Warshall pour calculer
        # une "matrice-dictionnaire", qui a un couple d'etats associe
        # la longueur d'un plus court chemin allant du premier vers le deuxieme.
        self.accessibles={}
        fini={}
        for i in self.label :
        	self.accessibles[i]=set([])
        	fini[i]=False
        
        def explore(ans, s):
        	ans.add(s)
        	for s2 in self.succ[s] :
        		if fini[s2] :
        			ans=ans.union(self.accessibles[s2])
        		elif not (s2 in ans) :
        			ans=explore(ans, s2)
        	return ans
        for s in self.label :
        	self.accessibles[s]=explore(self.accessibles[s], s)
        	fini[s]=True
#            self.accessibles[i]={}
#            for j in self.label :
#                if j == i :
#                    self.accessibles[i][j]=0
#                else :
#                    self.accessibles[i][j]=inf # infini
#        for i in self.label :
#            for j in self.succ[i] :
#                if self.accessibles[i][j] > 1 :
#                    self.accessibles[i][j]=1
#        for k in self.label :
#            for i in self.label :
#                for j in self.label :
#                    if self.accessibles[i][j] > self.accessibles[i][k]+self.accessibles[k][j] :
#                        self.accessibles[i][j]=self.accessibles[i][k]+self.accessibles[k][j]

        # On supprime tous les inaccessibles.
        # Comme ca, la boucle "for s2 in self.accessibles[s]"
        # fera bien un parcours sur tous les accessibles de s.
#        for s in self.accessibles :
#            toPop=[]
#            for s2 in self.accessibles[s] :
#                if self.accessibles[s][s2] == inf :
#                    toPop.append(s2)
#            for s2 in toPop :
#                self.accessibles[s].pop(s2)
        if (FLAGBUNTIL == 0) : 
            self.BUntil = len(self.accessibles[self.start])
        else :
            self.BUntil = FLAGBUNTIL
        self.TBVU = math.ceil(math.log(self.BUntil+1,2))
        self.TBVQ = math.ceil(math.log(len(self.accessibles[self.start])+1,2))
        self.Laccinit=[]
        for s in self.accessibles[self.start] :
            self.Laccinit.append(s)
        
                        
    def __init__(self, start, label, act,FLAGBUNTIL =0) :
        self.act=act
        self.start=start
        self.label=label
#        if FLAGBUNTIL > 0 :
#            print("Bounded until... "+str(FLAGBUNTIL))
        self.update(FLAGBUNTIL)

####################### pour les Until avec BitVec sous Z3
    def BUntilD(self) :
        return self.BUntil
    def cBUntil(self) :
        return self.cDistance(self.BUntil)
    def cDistance(self,v) : 
        if self.TBVU % 4 == 0 :
            res='#'+(hex(v)[1:])
            if len (res) < self.TBVU+2 :
                res = '#x'+('0'*((self.TBVU//4)+2-len(res)))+res[2:]
            return res 
        else :
            res='#'+(bin(v)[1:])
            if len (res) < self.TBVU+2 :
                res = '#b'+('0'*(self.TBVU+2-len(res)))+res[2:]
            return res
    def Numero(self,s) :
        if s in self.accessibles[self.start] :
            return self.Laccinit.index(s)
        else :
            return len(self.accessibles[self.start])
    def cNumero(self,s =None) :
        if s==None :
            ns=len(self.accessibles[self.start])
        else : 
            ns = self.Numero(s)
        if self.TBVQ % 4 == 0 :
            res='#'+(hex(ns)[1:])
            if len (res) < (self.TBVQ)/4+2 :
                res = '#x'+('0'*((self.TBVQ//4)+2-len(res)))+res[2:]
            return res 
        else :
            res='#'+(bin(ns)[1:])
            if len (res) < self.TBVQ+2 :
                res = '#b'+('0'*(self.TBVQ+2-len(res)))+res[2:]
            return res

#######################    
        
            
    # Enleve un etat sans appeler update.
    def remove(self, state) :
        if state == self.start :
            raise "Vous ne pouvez pas supprimer l'etat initial!"

        self.act.pop(state)
        self.label.pop(state)
        for s in self.label :
            for action in self.act[s] :
                for elt in self.act[s][action] :
                    if elt == state :
                        self.act[s][action].remove(state)
                    
    # Methode statique pour lire une structure a partir d'un fichier.
    def read(input,FLAGBUNTIL =0) :
        global start
        start=None
        label={}
        act={}

        class Situation(Enum) :
            DEFAULT=1
            IN_LABEL=2
            IN_SUCC=3
            IN_ACTION=4
            IN_STATES=5
            IN_TRANSITIONS=6
        
        global ident # Le mot qu'on est en train de lire.
        ident=""
        
        situation=Situation.DEFAULT # A quoi correspond ident?
        # IN_LABEL : une proposition atomique.
        # IN_SUCC : un successeur.
        # IN_ACTION : une action.
        # IN_STATES : un etat dont on va decrire le label.
        # IN_TRANSITIONS : un etat dont on va decrire les successeurs.
        
        global first # C'est le premier etat qu'on examine.
        first=True
        global currentState # De quel etat on regarde les successeurs ou les propositions atomiques.
        currentState=None
        global currentAction # Pour quelle action on examine les successeurs de currentState.
        currentAction=None

        def addState(st) :
            global currentState
            global currentAction
            global first
            global start
            currentState=st,
            if not ((st,) in label) : # Premiere declaration de st.
                label[st,]=set([])
                act[st,]={}
            if first :
                first=False
                start=st,
            currentAction=None
            pass

        def addAction(st, ac) :
            global currentAction
            if ac == "" :
                currentAction=None
            else :
                currentAction=ac
            if not (currentAction in act[st]) :
                act[st][currentAction]=set([])

        # La lecture d'un mot est finie, que doit on en faire?
        def endWord() :
            global ident
            if situation == Situation.IN_ACTION :
                addAction(currentState, ident)
            if ident == "" :
                return
            if situation == Situation.IN_STATES or situation == Situation.IN_TRANSITIONS :
                addState(ident)
            elif situation == Situation.IN_LABEL :
                label[currentState].add(ident)
            elif situation == Situation.IN_SUCC :
                act[currentState][currentAction].add((ident,))
            elif situation == Situation.DEFAULT :
                print("warning : states ou transitions non declare.\nLigne ignoree.")
            ident=""

        for line in input :
            ident=""
            if line == "states\n" :
                situation=Situation.IN_STATES
                continue
            if line == "transitions\n" :
                situation=Situation.IN_TRANSITIONS
                continue
            if line == "end\n" :
                break
            for car in line :
                if car == "|" or car.isspace():
                    endWord()
                    continue
                if car == "," :
                    endWord()
                    if situation == Situation.IN_SUCC :
                        situation = Situation.IN_TRANSITIONS
                    continue
                if car == "{" or car == ":" :
                    endWord()
                    situation=Situation.IN_LABEL
                    continue
                if car == "}" :
                    endWord()
                    situation=Situation.IN_STATES
                    continue
                if car == "-" :
                    endWord()
                    situation=Situation.IN_ACTION
                    continue
                if car == ">" :
                    endWord()
                    situation=Situation.IN_SUCC
                    continue
                ident=ident+car
            endWord()
            if situation == Situation.IN_SUCC :
                situation=Situation.IN_TRANSITIONS
            elif situation == Situation.IN_LABEL :
                situation=Situation.IN_STATES

        return Krip(start, label, act,FLAGBUNTIL)

    def from_file(filename,FLAGBUNTIL =0) :
        input=open(filename, "r")
        ans=Krip.read(input,FLAGBUNTIL)
        input.close()
        return ans

    # Il est important de ne pas oublier la ligne "end".
    def from_std() :
        return Krip.read(sys.stdin)

    # Les etats sont maintenant des tuples.
    def string_of_state(st) :
        ans=st[0]
        for c in st[1:] :
            ans=ans+"_"+c
        return ans

    # Renvoie le produit de synchronisation de krip1 et krip2.
    # Comme cette methode peut etre appelee plusieurs fois avant
    # qu'on travaille sur le produit final,
    # elle renvoie une structure qui est partiellement remplie,
    # ou update n'a pas ete appelee.
    # Pour avoir des structures de Kripke completes,
    # il faut utiliser multisync ou sync plus bas.
    def syncAux(krip1, krip2) :
        start=krip1.start+krip2.start
        firstLabel={}
        firstAct={}
        # Les actions en commun aux deux structures.
        actionsCommon=set([])
        
        for s1 in krip1.label :
            for s2 in krip2.label :
                new=s1+s2
                firstLabel[new]=krip1.label[s1].union(krip2.label[s2])
                firstAct[new]={}
                for action1 in krip1.act[s1] :
                    if (action1 == None) : # Il n'y a que krip1 qui bouge.
                        isDef=(None in firstAct[new])
                        for succ in krip1.act[s1][None] :
                            if not isDef :
                                isDef=True
                                firstAct[new][None]=set([])
                            firstAct[new][None].add(succ+s2)

                    else :
                        for action2 in krip2.act[s2] :
                            if action1 == action2 : # krip1 et krip2 bougent en meme temps.
                                actionsCommon.add(action1)
                                isDef=(action1 in firstAct[new])
                                for succ1 in krip1.act[s1][action1] :
                                    for succ2 in krip2.act[s2][action2] :
                                        if not isDef :
                                            isDef=True
                                            firstAct[new][action1]=set([])
                                        firstAct[new][action1].add(succ1+succ2)

                if None in krip2.act[s2] : # Il n'y a que krip2 qui bouge.
                    isDef=(None in firstAct[new])
                    for succ in krip2.act[s2][None] :
                        if not isDef :
                            isDef=True
                            firstAct[new][None]=set([])
                        firstAct[new][None].add(s1+succ)

        for s1 in krip1.label :
            for action in krip1.act[s1] :
                if not ((action == None) or (action in actionsCommon)) : # Seulement krip1 bouge.
                    for s2 in krip2.label :
                        new=s1+s2
                        if not (new in firstAct) :
                            firstAct[new]={}
                        firstAct[new][action]=set([])
                        for succ in krip1.act[s1][action] :
                            firstAct[new][action].add(succ+s2)

        for s2 in krip2.label :
            for action in krip2.act[s2] :
                if not ((action == None) or (action in actionsCommon)) : # Seulement krip2 bouge.
                    for s1 in krip1.label :
                        new=s1+s2
                        if not (new in firstAct) :
                            firstAct[new]={}
                        firstAct[new][action]=set([])
                        for succ in krip2.act[s2][action] :
                            firstAct[new][action].add(s1+succ)

        # Comme dans d'autres methodes, on a calcule trop d'etats,
        # dont certains sont inaccessibles.
        # On fait un parcours en profondeur pour se restreindre aux accessibles.
        marqued=set([])
        finalLabel={}
        finalAct={}
        def xPlore(state) :
            marqued.add(state)
            finalLabel[state]=firstLabel[state]
            finalAct[state]={}
            for action in firstAct[state] :
                finalAct[state][action]=firstAct[state][action]
                for succ in firstAct[state][action] :
                    if not (succ in marqued) :
                        xPlore(succ)

        xPlore(start)

        class PartialKrip :
            def __init__(self, start, label, act) : # On n'appelle pas update.
                self.start=start
                self.label=label
                self.act=act
                
        return PartialKrip(start, finalLabel, finalAct)

    def multiSync(kripList) :
        length=len(kripList)
        if length == 0 :
            raise "Liste vide dans la synchronisation."
        if length == 1 :
            return kripList[0]
        ans=kripList[0]
        for i in range(length-1) :
            ans=Krip.syncAux(ans, kripList[i+1])
        return Krip(ans.start, ans.label, ans.act)

    def sync(krip1, krip2) :
        return Krip.multiSync([krip1, krip2])
